// (c) mlinder
using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace HidApi
{
				public static class Hid
				{
								/// <summary>
								/// Hid device info.
								/// </summary>
								public class HidDeviceInfo
								{
												public ushort VendorID { get; private set; }

												public ushort ProductID { get; private set; }

												public ushort ReleaseNumber { get; private set; }

												public string Path { get; private set; }

												public string Manufacturer { get; private set; }

												public string Product { get; private set; }

												public string SerialNumber { get; private set; }

												public ushort UsagePage { get; private set; }

												public ushort Usage { get; private set; }

												public int InterfaceNumber { get; private set; }

												internal HidDeviceInfo (Native.hid_device_info info)
												{
																VendorID = info.vendor_id;
																ProductID = info.product_id;
																ReleaseNumber = info.release_number;
																Path = info.path;
																Manufacturer = Native.readStringUTF32 (info.manufacturer_string);
																Product = Native.readStringUTF32 (info.product_string);
																SerialNumber = Native.readStringUTF32 (info.serial_number);
																UsagePage = info.usage_page;
																Usage = info.usage;
																InterfaceNumber = info.interface_number;
												}
								}

	     
								/// <summary>
								/// HID Device
								/// </summary>
								public class HidDevice
								{
												/// <summary>
												/// Gets the handle.
												/// </summary>
												/// <value>
												/// The handle.
												/// </value>
												public IntPtr Handle { get; private set; }

												public HidDevice (IntPtr handle)
												{
																Handle = handle;
												}

												/// <summary>
												/// Sends the feature report.
												/// </summary>
												/// <param name='buf'>
												/// Buffer.
												/// </param>
												public void SendFeatureReport(byte[] buf)
												{
																if (!Native.hid_send_feature_report (Handle, buf, buf.Length))
																				throw new IOException ();
												}

												/// <summary>
												/// Gets the feature report.
												/// </summary>
												/// <param name='buf'>
												/// Buffer.
												/// </param>
												public void ReceiveFeatureReport(byte[] buf)
												{
																if (!Native.hid_get_feature_report (Handle, buf, buf.Length))
																				throw new IOException ();
												}

												/// <summary>
												/// Gets the feature report.
												/// </summary>
												/// <returns>
												/// The feature report.
												/// </returns>
												/// <param name='len'>
												/// Length.
												/// </param>
												public byte[] ReceiveFeatureReport(int len)
												{
																byte[] buf = new byte[len];
																ReceiveFeatureReport (buf);
																return buf;
												}
								}

								internal static class Native
								{
												public static unsafe string readStringUTF32(IntPtr ptr)
												{
																if (ptr == IntPtr.Zero)
																				return null;
																string str = "";
																byte* arr = (byte*)ptr;
																int i = 0;
																do {
																				ushort val = (ushort)(arr [i] | ((ushort)arr [i + 1] << 8));
																				if (val == 0)
																								break;
																				str += (char)val;
																				i += 4;
																} while (true);
																return str;
												}

												[StructLayout(LayoutKind.Sequential)]
												public class hid_device_info
												{
																public string path;
																public ushort vendor_id;
																public ushort product_id;
																public IntPtr serial_number;
																public ushort release_number;
																public IntPtr manufacturer_string;
																public IntPtr product_string;
																public ushort usage_page;
																public ushort usage;
																public int interface_number;
																private IntPtr _next;

																public hid_device_info next()
																{
																				if (_next == IntPtr.Zero)
																								return null;
																				return (hid_device_info)Marshal.PtrToStructure (_next, typeof(hid_device_info));
																}
												}

												[DllImport("libhidapi-libusb")]
												public static extern hid_device_info hid_enumerate(ushort vendor_id, ushort product_id);

												[DllImport("libhidapi-libusb")]
												public static extern IntPtr hid_open(ushort vendor_id, ushort product_id, IntPtr serial);

												[DllImport("libhidapi-libusb")]
												public static extern bool hid_send_feature_report(IntPtr handle, byte[] buf, int len);

												[DllImport("libhidapi-libusb")]
												public static extern bool hid_get_feature_report(IntPtr handle, byte[] buf, int len);
								}

								/// <summary>
								/// Enumerate the specified vendorId and productId.
								/// </summary>
								/// <param name='vendorId'>
								/// Vendor identifier.
								/// </param>
								/// <param name='productId'>
								/// Product identifier.
								/// </param>
								public static IEnumerable<HidDeviceInfo> Enumerate(ushort vendorId = 0x00, ushort productId = 0x00)
								{
												var info = Native.hid_enumerate (vendorId, productId);

												do {
																yield return new HidDeviceInfo (info);
																info = info.next ();
												} while (info != null);
								}

								/// <summary>
								/// Open the specified vendorId, productId and serialNumber.
								/// </summary>
								/// <param name='vendorId'>
								/// Vendor identifier.
								/// </param>
								/// <param name='productId'>
								/// Product identifier.
								/// </param>
								/// <param name='serialNumber'>
								/// Serial number.
								/// </param>
								public static HidDevice Open(ushort vendorId, ushort productId, string serialNumber = null)
								{
												IntPtr serialHandle = IntPtr.Zero;
												if (serialNumber != null) {
																byte[] utf32 = Encoding.UTF32.GetBytes (serialNumber + "\0");
																serialHandle = Marshal.AllocHGlobal (utf32.Length);
																Marshal.Copy (utf32, 0, serialHandle, utf32.Length);
												}

												IntPtr handle = Native.hid_open (vendorId, productId, serialHandle);
												return handle == IntPtr.Zero ? null : new HidDevice (handle);
								}
				}
}
